/* Copyright (C) 2016-2018, 2021 |Meso|Star> (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#include "scpr.h"

#include <rsys/double2.h>
#include <rsys/dynamic_array_double.h>
#include <rsys/dynamic_array_size_t.h>
#include <rsys/hash_table.h>
#include <rsys/mem_allocator.h>
#include <rsys/ref_count.h>

#include <polygon.h>
#include <clipper.hpp>
STATIC_ASSERT(sizeof(ClipperLib::cInt) >= sizeof(double), Unexpected_Type_Size);

struct vertex { double pos[2]; };
static FINLINE int
vertex_eq(const struct vertex* a, const struct vertex* b)
{ return d2_eq(a->pos, b->pos); }

/* Define the vertex to index hash table */
#define HTABLE_NAME vertex
#define HTABLE_DATA size_t
#define HTABLE_KEY struct vertex
#define HTABLE_KEY_FUNCTOR_EQ vertex_eq
#include <rsys/hash_table.h>

struct poly {
  struct darray_double coords;
  double lower[2], upper[2]; /* Polygon AABB */
};

struct scpr_mesh {
  struct darray_double coords;
  struct darray_size_t indices;

  ref_T ref;
  struct mem_allocator* allocator;
};

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static FINLINE ClipperLib::cInt
double_to_cInt(const double d, const double scale)
{
  double dbl = d;
  union { int64_t i; double d; } ucast;
  ClipperLib::cInt i;
  ASSERT(scale > 0);

 /* Map 'd' in [1, 2] => Fix the exponent */
  ucast.d = 1 + fabs(d / scale);

  /* Store the positive or null exponent in the 52^th bit. This ensure that if
   * 'd' is equal to 2 then it is not encoded as 0. */
  i = (((ucast.i >> 52) & 0x7FF) - 1023) << 52;

  /* Store the mantissa in the [0 .. 51] bits */
  i = (ucast.i & 0x000FFFFFFFFFFFFF) | i;

  /* Apply the sign to the resulting integer */
  return dbl < 0 ? -i : i;
}

static FINLINE double
cInt_to_double(const ClipperLib::cInt i, const double scale)
{
  double dbl;
  union { int64_t i; double d; } ucast;
  ASSERT(scale > 0);

  ucast.i = llabs(i);
  ucast.i = ((1023 + (ucast.i >> 52)) << 52) | (ucast.i & 0x000FFFFFFFFFFFFF);
  dbl = (ucast.d - 1) * scale;
  return i < 0 ? -dbl : dbl;
}

static FINLINE ClipperLib::ClipType
scpr_operation_to_clip_type(const enum scpr_operation op)
{
  ClipperLib::ClipType ctype = ClipperLib::ctIntersection;
  switch(op) {
    case SCPR_AND: ctype = ClipperLib::ctIntersection; break;
    case SCPR_SUB: ctype = ClipperLib::ctDifference; break;
    default: FATAL("Unreachable code\n"); break;
  }
  return ctype;
}

static INLINE int
aabb_intersect
  (const double lower0[2],
   const double upper0[2],
   const double lower1[2],
   const double upper1[2])
{
  ASSERT(lower0 && upper0 && lower1 && upper1);
  return lower0[0] < upper1[0]
      && lower0[1] < upper1[1]
      && lower1[0] < upper0[0]
      && lower1[1] < upper0[1];
}

static INLINE int
aabb_is_degenerated(const double lower[2], const double upper[2])
{
  ASSERT(lower && upper);
  return lower[0] >= upper[0] || lower[1] >= upper[1];
}

static INLINE void
poly_init(struct mem_allocator* allocator, struct poly* poly)
{
  ASSERT(allocator && poly);
  darray_double_init(allocator, &poly->coords);
}

static INLINE void
poly_release(struct poly* poly)
{
  ASSERT(poly);
  darray_double_release(&poly->coords);
}

static res_T
poly_setup(struct poly* poly, const struct scpr_polygon* desc)
{
  size_t ivert;
  res_T res = RES_OK;
  ASSERT(poly && desc);

  if(!desc->get_position && !desc->nvertices) {
    res = RES_BAD_ARG;
    goto error;
  }

  res = darray_double_resize(&poly->coords, desc->nvertices * 2/*#coords*/);
  if(res != RES_OK) goto error;

  d2_splat(poly->lower, DBL_MAX);
  d2_splat(poly->upper,-DBL_MAX);
  FOR_EACH(ivert, 0, desc->nvertices) {
    double* pos = darray_double_data_get(&poly->coords) + ivert*2;
    desc->get_position(ivert, pos, desc->context);
    d2_min(poly->lower, poly->lower, pos);
    d2_max(poly->upper, poly->upper, pos);
  }

exit:
  return res;
error:
  darray_double_clear(&poly->coords);
  d2_splat(poly->lower, DBL_MAX);
  d2_splat(poly->upper,-DBL_MAX);
  goto exit;
}

static void
triangle_compute_aabb
  (double tri[3][2],
   double lower[2],
   double upper[2])
{
  size_t ivert;
  ASSERT(tri && lower && upper);

  d2_splat(lower, DBL_MAX);
  d2_splat(upper,-DBL_MAX);
  FOR_EACH(ivert, 0, 3) {
    d2_min(lower, lower, tri[ivert]);
    d2_max(upper, upper, tri[ivert]);
  }
}

static res_T
register_vertex
  (const double pos[2],
   struct darray_double* coords,
   struct darray_size_t* indices,
   struct htable_vertex* vertices)
{
  struct vertex v;
  size_t* pid, id;
  res_T res = RES_OK;
  ASSERT(pos && coords && indices && vertices);

  d2_set(v.pos, pos);

  /* FIXME dirty hack */
  v.pos[0] = (float)v.pos[0];
  v.pos[1] = (float)v.pos[1];

  pid = htable_vertex_find(vertices, &v);

  if(pid) {
    id = *pid;
  } else {
    const size_t ivert = darray_double_size_get(coords);

    res = darray_double_resize(coords, ivert+2/*#coords*/);
    if(res != RES_OK) goto error;
    d2_set(darray_double_data_get(coords)+ivert, pos);

    id = ivert / 2;
    res = htable_vertex_set(vertices, &v, &id);
    if(res != RES_OK) goto error;
  }

  res = darray_size_t_push_back(indices, &id);
  if(res != RES_OK) goto error;

exit:
  return res;
error:
  goto exit;
}

static FINLINE res_T
register_triangle
  (double tri[3][2],
   struct darray_double* coords, /* Vertex buffer */
   struct darray_size_t* indices, /* Index buffer */
   struct htable_vertex* vertices) /* Map a vertex to its index */
{
  size_t ivert;
  res_T res = RES_OK;
  ASSERT(tri && coords && indices && vertices);
  FOR_EACH(ivert, 0, 3) {
    res = register_vertex(tri[ivert], coords, indices, vertices);
    if(res != RES_OK) return res;
  }
  return RES_OK;
}


static res_T
register_paths
  (const ClipperLib::Paths& paths,
   struct darray_double* coords, /* Vertex buffer */
   struct darray_size_t* indices, /* Index buffer */
   struct htable_vertex* vertices, /* Map a vertex to its index */
   const double extend[2], /* Scale to apply to the cInt coordinates */
   struct polygon* polygon) /* Use to triangulate the clipped polygons */
{
  size_t ivert;
  size_t ipath;
  res_T res = RES_OK;
  ASSERT(coords && indices && vertices);

  FOR_EACH(ipath, 0, paths.size()) {
    if(paths[ipath].size() == 3) {
      FOR_EACH(ivert, 0, 3) {
        double pos[2];
        pos[0] = cInt_to_double(paths[ipath][ivert].X, extend[0]);
        pos[1] = cInt_to_double(paths[ipath][ivert].Y, extend[1]);
        res = register_vertex(pos, coords, indices, vertices);
        if(res != RES_OK) goto error;
      }
    } else {
      /* Triangulate the clipped primitive */
      const uint32_t* ids;
      uint32_t nids;

      /* Define the contour of the polygon to triangulate */
      POLYGON(clear(polygon));
      FOR_EACH(ivert, 0, paths[ipath].size()) {
        float fpos[3] = {0.f, 0.f, 0.f};
        double pos[2];

        pos[0] = cInt_to_double(paths[ipath][ivert].X, extend[0]);
        pos[1] = cInt_to_double(paths[ipath][ivert].Y, extend[1]);

        fpos[0] = (float)pos[0], fpos[1] = (float)pos[1];
        res = polygon_vertex_add(polygon, fpos);
        if(res != RES_OK) goto error;
      }
      res = polygon_triangulate(polygon, &ids, &nids);
      if(res != RES_OK) goto error;

      FOR_EACH(ivert, 0, nids) {
        float fpos[3];
        double pos[2];
        POLYGON(vertex_get(polygon, ids[ivert], fpos));
        pos[0] = (float)fpos[0];
        pos[1] = (float)fpos[1];
        res = register_vertex(pos, coords, indices, vertices);
        if(res != RES_OK) goto error;
      }
    }
  }
exit:
  return res;
error:
  goto exit;
}

static void
mesh_compute_aabb(const struct scpr_mesh* mesh, double lower[2], double upper[2])
{
  size_t itri, ntris;

  SCPR(mesh_get_triangles_count(mesh, &ntris));
  d2_splat(lower, DBL_MAX);
  d2_splat(upper,-DBL_MAX);

  FOR_EACH(itri, 0, ntris) {
    size_t ids[3], ivert;
    SCPR(mesh_get_indices(mesh, itri, ids));
    FOR_EACH(ivert, 0, 3) {
      double pos[2];
      SCPR(mesh_get_position(mesh, ids[ivert], pos));
      d2_min(lower, lower, pos);
      d2_max(upper, upper, pos);
    }
  }
}

static void
mesh_release(ref_T* ref)
{
  struct scpr_mesh* mesh;
  ASSERT(ref);
  mesh = CONTAINER_OF(ref, struct scpr_mesh, ref);
  darray_double_release(&mesh->coords);
  darray_size_t_release(&mesh->indices);
  MEM_RM(mesh->allocator, mesh);
}

/*******************************************************************************
 * Exported functions
 ******************************************************************************/
res_T
scpr_mesh_create(struct mem_allocator* mem_allocator, struct scpr_mesh** out_mesh)
{
  struct scpr_mesh* mesh = NULL;
  struct mem_allocator* allocator = NULL;
  res_T res = RES_OK;

  if(!out_mesh) {
    res = RES_BAD_ARG;
    goto error;
  }

  allocator = mem_allocator ? mem_allocator : &mem_default_allocator;
  mesh = (struct scpr_mesh*)
    MEM_CALLOC(allocator, 1, sizeof(struct scpr_mesh));
  if(!mesh) {
    res = RES_MEM_ERR;
    goto error;
  }
  darray_double_init(mesh->allocator, &mesh->coords);
  darray_size_t_init(mesh->allocator, &mesh->indices);
  ref_init(&mesh->ref);
  mesh->allocator = allocator;

exit:
  if(out_mesh) *out_mesh = mesh;
  return res;

error:
  if(mesh) {
    SCPR(mesh_ref_put(mesh));
    mesh = NULL;
  }
  goto exit;
}

res_T
scpr_mesh_ref_get(struct scpr_mesh* mesh)
{
  if(!mesh) return RES_BAD_ARG;
  ref_get(&mesh->ref);
  return RES_OK;
}

res_T
scpr_mesh_ref_put(struct scpr_mesh* mesh)
{
  if(!mesh) return RES_BAD_ARG;
  ref_put(&mesh->ref, mesh_release);
  return RES_OK;
}

res_T
scpr_mesh_setup_indexed_vertices
  (struct scpr_mesh* mesh,
   const size_t ntris,
   void (*get_indices)(const size_t itri, size_t ids[3], void* ctx),
   const size_t nverts,
   void (*get_position)(const size_t ivert, double pos[2], void* ctx),
   void* data)
{
  double* pos = NULL;
  size_t* ids = NULL;
  size_t i;
  res_T res = RES_OK;

  if(!mesh || !ntris || !get_indices || !nverts || !get_position) {
    res = RES_BAD_ARG;
    goto error;
  }

  res = darray_double_resize(&mesh->coords, nverts*2/*#coords per vertex*/);
  if(res != RES_OK) goto error;
  res = darray_size_t_resize(&mesh->indices, ntris*3/*#vertices per triangle*/);
  if(res != RES_OK) goto error;

  /* Fetch mesh positions */
  pos = darray_double_data_get(&mesh->coords);
  FOR_EACH(i, 0, nverts) {
    get_position(i, pos+i*2, data);
  }

  /* Fetch mesh indices */
  ids = darray_size_t_data_get(&mesh->indices);
  FOR_EACH(i, 0, ntris) {
    get_indices(i, ids +i*3, data);
    if(ids[i*3+0] >= nverts
    || ids[i*3+1] >= nverts
    || ids[i*3+2] >= nverts) {
      res = RES_BAD_ARG;
      goto error;
    }
  }

exit:
  return res;
error:
  if(mesh) {
    darray_double_clear(&mesh->coords);
    darray_size_t_clear(&mesh->indices);
  }
  goto exit;
}

res_T
scpr_mesh_get_vertices_count(const struct scpr_mesh* mesh, size_t* nverts)
{
  if(!mesh || !nverts) return RES_BAD_ARG;
  ASSERT((darray_double_size_get(&mesh->coords) % 2/*#coords per vertex*/)==0);
  *nverts = darray_double_size_get(&mesh->coords) / 2/*#coords per vertex*/;
  return RES_OK;
}

res_T
scpr_mesh_get_triangles_count(const struct scpr_mesh* mesh, size_t* ntris)
{
  if(!mesh || !ntris) return RES_BAD_ARG;
  ASSERT((darray_size_t_size_get(&mesh->indices)%3/*#vertices per triangle*/)==0);
  *ntris = darray_size_t_size_get(&mesh->indices) / 3/*#vertices per triangle*/;
  return RES_OK;
}

res_T
scpr_mesh_get_indices
  (const struct scpr_mesh* mesh, const size_t itri, size_t ids[3])
{
  size_t ntris;
  const size_t i = itri * 3/*#vertices per triangle*/;
  if(!mesh || !ids) return RES_BAD_ARG;
  SCPR(mesh_get_triangles_count(mesh, &ntris));
  if(itri >= ntris) return RES_BAD_ARG;
  ids[0] = darray_size_t_cdata_get(&mesh->indices)[i+0];
  ids[1] = darray_size_t_cdata_get(&mesh->indices)[i+1];
  ids[2] = darray_size_t_cdata_get(&mesh->indices)[i+2];
  return RES_OK;
}

res_T
scpr_mesh_get_position
  (const struct scpr_mesh* mesh, const size_t ivert, double pos[2])
{
  size_t nverts;
  const size_t i = ivert * 2/*#coords per vertex*/;
  if(!mesh || !pos) return RES_BAD_ARG;
  SCPR(mesh_get_vertices_count(mesh, &nverts));
  if(ivert >= nverts) return RES_BAD_ARG;
  pos[0] = darray_double_cdata_get(&mesh->coords)[i+0];
  pos[1] = darray_double_cdata_get(&mesh->coords)[i+1];
  return RES_OK;
}

res_T
scpr_mesh_clip
  (struct scpr_mesh* mesh,
   const enum scpr_operation op,
   struct scpr_polygon* poly_desc)
{
  double lower[2], upper[2], extend[2];
  struct poly poly;
  struct polygon* polygon = NULL; /* Use to triangulate clipped polygons */
  struct darray_double coords; /* Coordinates of the clipped mesh */
  struct darray_size_t indices; /* Indices of the clipped mesh */
  struct htable_vertex vertices; /* Map a coordinate to its index */
  ClipperLib::Clipper clipper(ClipperLib::ioStrictlySimple);
  ClipperLib::Paths output; /* Contour of the clipped polgyon */
  ClipperLib::Path cand_path; /* Contour of the candidate polygon */
  ClipperLib::Path clip_path; /* Contour of the clip polygon */
  ClipperLib::ClipType clip_type; /* Type of clipping to perform */
  size_t ivert, nverts;
  size_t itri, ntris;

  res_T res = RES_OK;

  if(!mesh || !poly_desc || (unsigned)op >= SCPR_OPERATIONS_COUNT__)
    return RES_BAD_ARG;

  clip_type = scpr_operation_to_clip_type(op);

  darray_double_init(mesh->allocator, &coords);
  darray_size_t_init(mesh->allocator, &indices);
  htable_vertex_init(mesh->allocator, &vertices);

  poly_init(mesh->allocator, &poly);
  res = poly_setup(&poly, poly_desc);
  if(res != RES_OK) goto error;
  if(aabb_is_degenerated(poly.lower, poly.upper)) goto exit;

  mesh_compute_aabb(mesh, lower, upper);
  if(aabb_is_degenerated(lower, upper)) goto exit;

  /* Compute the overall aabb of the candidate and the clip polygon */
  d2_min(lower, lower, poly.lower);
  d2_max(upper, upper, poly.upper);
  d2_sub(extend, upper, lower);

  /* Setup the clip path */
  nverts = darray_double_size_get(&poly.coords) / 2/*#coords per vertex*/;
  FOR_EACH(ivert, 0, nverts) {
    const double* v = darray_double_cdata_get(&poly.coords) + ivert*2;
    ClipperLib::IntPoint pt;
    pt.X = double_to_cInt(v[0], extend[0]);
    pt.Y = double_to_cInt(v[1], extend[1]);
    clip_path.push_back(pt);
  }

  /* Create the polygon structure used to triangulate the clipped polygons */
  res = polygon_create(mesh->allocator, &polygon);
  if(res != RES_OK) goto error;

  /* Clip the triangles of the mesh */
  SCPR(mesh_get_triangles_count(mesh, &ntris));
  FOR_EACH(itri, 0, ntris) {
    size_t ids[3];
    double tri[3][2];

    /* Fetch the triangle vertices and compute its AABB */
    SCPR(mesh_get_indices(mesh, itri, ids));
    SCPR(mesh_get_position(mesh, ids[0], tri[0]));
    SCPR(mesh_get_position(mesh, ids[1], tri[1]));
    SCPR(mesh_get_position(mesh, ids[2], tri[2]));
    triangle_compute_aabb(tri, lower, upper);

    /* Do not clip triangles that don not intersect the clip AABB */
    if(!aabb_intersect(lower, upper, poly.lower, poly.upper)) {
      if(op != SCPR_AND) {
        res = register_triangle(tri, &coords, &indices, &vertices);
        if(res != RES_OK) goto error;
      }
      continue;
    }

    /* Setup the candidate path */
    cand_path.clear();
    FOR_EACH(ivert, 0, 3) {
      ClipperLib::IntPoint pt;
      pt.X = double_to_cInt(tri[ivert][0], extend[0]);
      pt.Y = double_to_cInt(tri[ivert][1], extend[1]);
      cand_path.push_back(pt);
    }

    /* Clip the polygon */
    clipper.Clear();
    clipper.AddPath(cand_path, ClipperLib::ptSubject, 1);
    clipper.AddPath(clip_path, ClipperLib::ptClip, 1);
    clipper.Execute(clip_type, output);

    /* Register the resulting clipped polygons */
    res = register_paths
      (output, &coords, &indices, &vertices, extend, polygon);
    if(res != RES_OK) goto error;
  }

  res = darray_double_copy_and_clear(&mesh->coords, &coords);
  if(res != RES_OK) FATAL("Unexpected error.\n");
  res = darray_size_t_copy_and_clear(&mesh->indices, &indices);
  if(res != RES_OK) FATAL("Unexpected error.\n");

exit:
  if(polygon) POLYGON(ref_put(polygon));
  darray_double_release(&coords);
  darray_size_t_release(&indices);
  htable_vertex_release(&vertices);
  poly_release(&poly);
  return res;
error:
  goto exit;
}

