/* Copyright (C) 2015, 2016, 2019, 2021 |Meso|Star> (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "sstl.h"
#include "test_sstl_utils.h"

#include <rsys/clock_time.h>
#include <rsys/float3.h>
#include <rsys/logger.h>

static void
test_basic(struct sstl* sstl)
{
  static const char* test0 =
    "solid\n"
    "  facet normal 0.0 -1.0 0.0\n"
    "    outer loop\n"
    "      vertex 0.0 0.0 0.0\n"
    "      vertex 1.0 0.0 0.0\n"
    "      vertex 0.0 0.0 1.0\n"
    "    endloop\n"
    "  endfacet\n"
    "endsolid";
  static const char* test1 =
    "solid my_solid\n"
    "\n"
    "  facet normal 0.0 -1.0 0.0\n"
    "    outer loop     hophophophophop\n"
    "      vertex\t  0.0      0.0 0.0\n"
    "      vertex 1.0   0.0 0.0     \taaa\n"
    "      vertex 0.0   0.0     1.0\n"
    "    endloop   \n"
    "  endfacet   \t\t\t noise\n"
    "endsolid pouet\n";
  static const char* test2 =
    "solid my_solid\n"
    "endsolid my_solid\n";
  static const char* test3 =
    "solid\n"
    "  facet normal 0.0 0.0 0.0\n"
    "    outer loop\n"
    "      vertex 0.0 0.0 0.0\n"
    "      vertex 1.0 0.0 0.0\n"
    "      vertex 0.0 0.0 1.0\n"
    "    endloop\n"
    "  endfacet\n"
    "endsolid";
  static const char* bad[] = {
    "solid\n"
    "  facet normal 0.0 -1.0 0.0\n"
    "    outer loop\n"
    "      vertex 0.0 0.0 0.0\n"
    "      vertex 1.0 0.0 0.0\n"
    "      vertex 0.0 0.0 1.0\n"
    "    endloop\n"
    "  endfacet\n"
    ,
    "  facet normal 0.0 -1.0 0.0\n"
    "    outer loop\n"
    "      vertex 0.0 0.0 0.0\n"
    "      vertex 1.0 0.0 0.0\n"
    "      vertex 0.0 0.0 1.0\n"
    "    endloop\n"
    "  endfacet\n"
    "endsolid\n"
    ,
    "solid\n"
    "  facet 0.0 -1.0 0.0\n"
    "    outer loop\n"
    "      vertex 0.0 0.0 0.0\n"
    "      vertex 1.0 0.0 0.0\n"
    "      vertex 0.0 0.0 1.0\n"
    "    endloop\n"
    "  endfacet\n"
    "endsolid\n"
    ,
    "solid\n"
    "  normal 0.0 -1.0 0.0\n"
    "    outer loop\n"
    "      vertex 0.0 0.0 0.0\n"
    "      vertex 1.0 0.0 0.0\n"
    "      vertex 0.0 0.0 1.0\n"
    "    endloop\n"
    "  endfacet\n"
    "endsolid\n"
    ,
    "solid\n"
    "  facet normal 0.0 -1.0 0.0\n"
    "    outer loop\n"
    "      vertex 0.0 0.0 a.0\n"
    "      vertex 1.0 0.0 0.0\n"
    "      vertex 0.0 0.0 1.0\n"
    "    endloop\n"
    "  endfacet\n"
    "endsolid\n"
    ,
    "solid\n"
    "  facet normal 0.0 -1.0 0.0\n"
    "    outer loop\n"
    "      vertex 1.0 0.0 0.0\n"
    "      vertex 0.0 0.0 1.0\n"
    "    endloop\n"
    "  endfacet\n"
    "endsolid\n"
    ,
    "solid\n"
    "  facet normal 0.0 -1.0 0.0\n"
    "    vertex 0.0 0.0 0.0\n"
    "    vertex 1.0 0.0 0.0\n"
    "    vertex 0.0 0.0 1.0\n"
    "  endfacet\n"
    "endsolid\n"
  };
  const size_t nbads = sizeof(bad)/sizeof(const char*);
  float tmp[3];
  struct sstl_desc desc;
  FILE* file;
  size_t i;

  CHK(sstl != NULL);

  file = fopen("test_basic.stl", "w");
  CHK(file != NULL);
  fwrite(test0, sizeof(char), strlen(test0), file);
  fclose(file);

  CHK(sstl_load(NULL, NULL) == RES_BAD_ARG);
  CHK(sstl_load(sstl, NULL) == RES_BAD_ARG);
  CHK(sstl_load(NULL, "test_basic.stl") == RES_BAD_ARG);
  CHK(sstl_load(sstl, "none.stl") == RES_IO_ERR);
  CHK(sstl_load(sstl, "test_basic.stl") == RES_OK);

  CHK(sstl_get_desc(NULL, NULL) == RES_BAD_ARG);
  CHK(sstl_get_desc(sstl, NULL) == RES_BAD_ARG);
  CHK(sstl_get_desc(NULL, &desc) == RES_BAD_ARG);
  CHK(sstl_get_desc(sstl, &desc) == RES_OK);

  CHK(desc.solid_name == NULL);
  CHK(desc.vertices_count == 3);
  CHK(desc.triangles_count == 1);
  CHK(desc.indices[0] == 0);
  CHK(desc.indices[1] == 1);
  CHK(desc.indices[2] == 2);
  CHK(f3_eq(desc.vertices + 0*3, f3(tmp, 0.f, 0.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + 1*3, f3(tmp, 1.f, 0.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + 2*3, f3(tmp, 0.f, 0.f, 1.f)) == 1);
  CHK(f3_eq(desc.normals, f3(tmp, 0.f, -1.f, 0.f)) == 1);

  file = tmpfile();
  CHK(file != NULL);
  fwrite(test1, sizeof(char), strlen(test1), file);
  rewind(file);
  CHK(sstl_load_stream(NULL, NULL) == RES_BAD_ARG);
  CHK(sstl_load_stream(sstl, NULL) == RES_BAD_ARG);
  CHK(sstl_load_stream(NULL, file) == RES_BAD_ARG);
  CHK(sstl_load_stream(sstl, file) == RES_OK);
  fclose(file);

  CHK(sstl_get_desc(sstl, &desc) == RES_OK);
  CHK(strcmp(desc.solid_name, "my_solid") == 0);
  CHK(desc.vertices_count == 3);
  CHK(desc.triangles_count == 1);
  CHK(desc.indices[0] == 0);
  CHK(desc.indices[1] == 1);
  CHK(desc.indices[2] == 2);
  CHK(f3_eq(desc.vertices + 0*3, f3(tmp, 0.f, 0.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + 1*3, f3(tmp, 1.f, 0.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + 2*3, f3(tmp, 0.f, 0.f, 1.f)) == 1);
  CHK(f3_eq(desc.normals, f3(tmp, 0.f, -1.f, 0.f)) == 1);

  file = tmpfile();
  fwrite(test2, sizeof(char), strlen(test2), file);
  rewind(file);
  CHK(sstl_load_stream(sstl, file) == RES_OK);
  fclose(file);

  CHK(sstl_get_desc(sstl, &desc) == RES_OK);
  CHK(strcmp(desc.solid_name, "my_solid") == 0);
  CHK(desc.vertices_count == 0);
  CHK(desc.triangles_count == 0);

  file = tmpfile();
  fwrite(test3, sizeof(char), strlen(test3), file);
  rewind(file);
  CHK(sstl_load_stream(sstl, file) == RES_OK);
  fclose(file);

  CHK(sstl_get_desc(sstl, &desc) == RES_OK);
  CHK(desc.vertices_count == 3);
  CHK(desc.triangles_count == 1);
  CHK(f3_eq(desc.normals, f3(tmp, 0.f, -1.f, 0.f)) == 1);

  FOR_EACH(i, 0, nbads) {
    file = tmpfile();
    fwrite(bad[i], sizeof(char), strlen(bad[i]), file);
    rewind(file);
    CHK(sstl_load_stream(sstl, file) == RES_BAD_ARG);
    fclose(file);
  }
}

static void
test_tetrahedron(struct sstl* sstl)
{
  static const char* tetrahedron[] = {
    "solid cube_corner\n",
    "  facet normal 0.0 -1.0 0.0\n",
    "    outer loop\n",
    "      vertex 0.0 0.0 0.0\n",
    "      vertex 1.0 0.0 0.0\n",
    "      vertex 0.0 0.0 1.0\n",
    "    endloop\n",
    "  endfacet\n",
    "  facet normal 0.0 0.0 -1.0\n",
    "    outer loop\n",
    "      vertex 0.0 0.0 0.0\n",
    "      vertex 0.0 1.0 0.0\n",
    "      vertex 1.0 0.0 0.0\n",
    "    endloop\n",
    "  endfacet\n",
    "  facet normal -1.0 0.0 0.0\n",
    "    outer loop\n",
    "      vertex 0.0 0.0 0.0\n",
    "      vertex 0.0 0.0 1.0\n",
    "      vertex 0.0 1.0 0.0\n",
    "    endloop\n",
    "  endfacet\n",
    "  facet normal 0.577 0.577 0.577\n",
    "    outer loop\n",
    "      vertex 1.0 0.0 0.0\n",
    "      vertex 0.0 1.0 0.0\n",
    "      vertex 0.0 0.0 1.0\n",
    "    endloop\n",
    "  endfacet\n",
    "endsolid\n"
  };
  FILE* file;
  const size_t nlines = sizeof(tetrahedron)/sizeof(const char*);
  struct sstl_desc desc;
  float tmp[3];
  size_t i;

  CHK(sstl != NULL);

  file = tmpfile();
  CHK(file != NULL);
  FOR_EACH(i, 0, nlines)
    fwrite(tetrahedron[i], sizeof(char), strlen(tetrahedron[i]), file);
  rewind(file);

  CHK(sstl_load_stream(sstl, file) == RES_OK);

  CHK(sstl_get_desc(sstl, &desc) == RES_OK);
  CHK(strcmp(desc.solid_name, "cube_corner") == 0);
  CHK(desc.vertices_count == 4);
  CHK(desc.triangles_count == 4);

  CHK(f3_eq(desc.vertices + desc.indices[0]*3, f3(tmp, 0.f, 0.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + desc.indices[1]*3, f3(tmp, 1.f, 0.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + desc.indices[2]*3, f3(tmp, 0.f, 0.f, 1.f)) == 1);
  CHK(f3_eq(desc.vertices + desc.indices[3]*3, f3(tmp, 0.f, 0.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + desc.indices[4]*3, f3(tmp, 0.f, 1.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + desc.indices[5]*3, f3(tmp, 1.f, 0.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + desc.indices[6]*3, f3(tmp, 0.f, 0.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + desc.indices[7]*3, f3(tmp, 0.f, 0.f, 1.f)) == 1);
  CHK(f3_eq(desc.vertices + desc.indices[8]*3, f3(tmp, 0.f, 1.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + desc.indices[9]*3, f3(tmp, 1.f, 0.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + desc.indices[10]*3, f3(tmp, 0.f, 1.f, 0.f)) == 1);
  CHK(f3_eq(desc.vertices + desc.indices[11]*3, f3(tmp, 0.f, 0.f, 1.f)) == 1);

  CHK(f3_eq(desc.normals + 0*3, f3(tmp, 0.f,-1.f, 0.f)) == 1);
  CHK(f3_eq(desc.normals + 1*3, f3(tmp, 0.f, 0.f,-1.f)) == 1);
  CHK(f3_eq(desc.normals + 2*3, f3(tmp,-1.f, 0.f, 0.f)) == 1);
  f3_normalize(tmp, f3(tmp, 1.f, 1.f, 1.f));
  CHK(f3_eq_eps(desc.normals + 3*3, tmp, 1.e-6f) == 1);
}

int
main(int argc, char** argv)
{
  struct mem_allocator allocator;
  struct sstl* sstl;
  int i;
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  CHK(sstl_create(NULL, &allocator, 1, &sstl) == RES_OK);

  test_basic(sstl);
  test_tetrahedron(sstl);

  FOR_EACH(i, 1, argc) {
    struct sstl_desc desc;
    char buf[512];
    struct time t0, t1;

    printf("loading %s", argv[i]);
    fflush(stdout);

    time_current(&t0);
    CHK(sstl_load(sstl, argv[i]) == RES_OK);
    time_current(&t1);
    time_sub(&t0, &t1, &t0);
    time_dump(&t0, TIME_MIN|TIME_SEC|TIME_MSEC, NULL, buf, sizeof(buf));

    CHK(sstl_get_desc(sstl, &desc) == RES_OK);

    printf(" - #vertices = %lu; #triangles = %lu - %s\n",
      (unsigned long)desc.vertices_count,
      (unsigned long)desc.triangles_count,
      buf);
  }

  CHK(sstl_ref_put(sstl) == RES_OK);

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHK(mem_allocated_size() == 0);
  return 0;
}

